# BLUE IRIS DAT FILE IMAGE EXTRACT UTILITY
# WORKS FOR ALL *.DAT FILES
# - DATABASE FILES, LIKE 'C:\BlueIris\db\clips.dat'
# - AI DAT FILES, LIKE 'D:\BlueIris\Alerts\DW2.20250124_150000.2814048.5-0.dat'
#
# KEY POINT: 
# AS A PRECAUTION, THIS SCRIPT ONLY USES A COPY OF ACTIVE DATABASE DAT FILES
# THESE FILES ARE AUTOMATICALLY COPIED TO THE SCRIPT'S WORKING FOLDER BEFORE USE
# 
# USAGE: {this_script.name).ps1 [args]
# ARGS:  -SourceFile  = (string) alternate path to a COPY of Blue Iris database file, e.g., "C:temp\clips.dat"
#                       NOTE: typically the database file is located at "C:\BlueIris\db\clips.dat"
#        -DestPath    = (string) alternate working path for script; default is "C:\temp"
#        -Extract     = (switch) extract JPG image(s) from the -SourceFile
#        -Range       = (int array) range of JPG images to be extracted from the -SourceFile
# RETURNS: [int] count of images in teh DAT file
# 
# Version: 1
# Created by jaydeel at ipcamtalk.com
#
# Updates:
# V1.0 - 2025-02-04 - initial release

param(
	[String]$SourceFile,
	[String]$DestPath,
	[Switch]$Extract,
	[Int[]]$Range,
	[Switch]$Help,
	[Switch]$DevMode
)

# For debugging
# Write-Host ("Range type = " + $Range.GetType().Name)
# Write-Host ("Range cnt = " + $Range.Count)
# Write-Host ("Range = " + ($Range | ConvertTo-Json))

$ScriptVersion = "V1"

# GATHER/VALIDATE SCRIPT & ENVIRONMENT INFO
$ScriptName = Split-Path -leaf $PSCommandpath
$ScriptBaseName = $ScriptName.Replace('.ps1', '')
# DETERMINE IF RUNNING IN BACKGROUND (IN CLI MODE, OR FROM TASK SCHEDULER, BLUE IRIS, etc.)
# NOTE: Scripts called by another script have a 'PSCommandPath'.
#       Method 'PSCommandPath' returns null if script is executed via the terminal
#       https://stackoverflow.com/a/23370618/13499984
$CallingScript = $MyInvocation.PSCommandPath
$FromConsole = $False  # Default
if ($null -eq $CallingScript) {
	$FromConsole = $True
} elseif ($CallingScript -match $ScriptName) {  # Detect recursive calls
	$FromConsole = $True
}
# For testing
# $FromConsole = $False

if ($Help -and $FromConsole) {
	Write-Host "========================================================================================================"
	Write-Host "BLUE IRIS *.DAT FILE IMAGE EXTRACT UTILITY"
	Write-Host "WORKS FOR ALL *.DAT FILES:"
	Write-Host "- DATABASE FILES, LIKE 'C:\BlueIris\db\clips.dat'"
	Write-Host "- AI DAT FILES, LIKE 'D:\BlueIris\Alerts\DW2.20250124_150000.2814048.5-0.dat'"
	Write-Host "--------------------------------------------------------------------------------------------------------"
	Write-Host "Usage:"
	Write-Host "   $ScriptName {no args} ... show this help guide"
	Write-Host "   $ScriptName -SourceFile 'C:\BlueIris\db\clips.dat' -Extract"
	Write-Host "   $ScriptName -SourceFile 'C:\BlueIris\db\clips.dat' -Extract -Range 1,5 ... first 5 images"
	Write-Host "   $ScriptName -SourceFile 'C:\BlueIris\db\clips.dat' -Extract -Range 5 ... ditto"
	Write-Host "   $ScriptName -SourceFile 'C:\BlueIris\db\clips.dat' -Extract -Range -5 ... last 5 images"
	Write-Host "   $ScriptName -SourceFile 'C:\BlueIris\db\clips.dat' -Extract -Range 0 ... all images"
	Write-Host "Args:"
	Write-Host "   -SourceFile ... full path to DAT file [1]"
	Write-Host "   -DestPath ... alternate root path to the working folder for this script [2]"
	Write-Host "   -Extract ... extract JPG image(s); multiple images requires -Range argument [3,4]"
	Write-Host "   -Range ... range of JPG images files to extract [5]"
	Write-Host "   -Help ... show this usage guide"
	Write-Host "   -DevMode ... show timing results"
	Write-Host "NOTES:"
	Write-Host "   [1] Like 'C:\BlueIris\db\clips.dat', or 'D:\BlueIris\Alerts\DW2.20250124_150000.2814048.5-0.dat'"
	Write-Host "       KEY POINT: Active BI database files are ALWAYS copied to the working folder before use"
	Write-Host "   [2] The default root path for the working folder is 'C:\temp'; it is created if necessary"
	Write-Host "   [3] The extraction folder is a subfolder in the working folder; '"
	Write-Host "       like 'C:\temp\$ScriptBaseName\extracted_jpgs'"
	Write-Host "   [4] Extracts only the first image only when used WITHOUT argument -Range;"
	Write-Host "       this is done to avoid the inadvertent extraction of thousands of images"
	Write-Host "   [5] Argument -Range accepts a comma delimited range like '-Range 1,5', "
	Write-Host "       OR a single UNSIGNED integer like '-Range N' to extract the first N images from -SourceFile"
	Write-Host "       OR a single NEGATIVE integer like '-Range -N' to extract the last N images from -SourceFile"
	Write-Host "       OR a zero '-Range 0' to extract ALL embedded images from -SourceFile"
	Write-Host "--------------------------------------------------------------------------------------------------------"
	exit	
}

# MAKE WORKING FOLDER
if ($DestPath) {
	$dest_path = $DestPath
} else {
	$dest_path = "C:\temp"
}
$dest_path = $dest_path + "\" + $ScriptBaseName
if (!(Test-Path -LiteralPath $dest_path)) {
	New-Item -ItemType Directory -Force -Path "$dest_path" | Out-Null
}

# MAKE FOLDER FOR EXTRACTED JPGS
$extract_path = $dest_path + "\extracted_jpgs"
if (!(Test-Path -LiteralPath $extract_path)) {
	New-Item -ItemType Directory -Force -Path "$extract_path" | Out-Null
}

# SPECIAL HANDLING FOR BLUE IRIS DATABASE FILES
$source_is_db_clips_dat_file = $false
$sfx = ''
if ($SourceFile) {  
	if (Test-Path -LiteralPath $SourceFile) {

		# Test if the requested *.dat file appears to be sourced from the Blue Iris \db\ folder
		if ($SourceFile.IndexOf('\db\') -gt 0) {
			
			# Flag a special case
			if ($SourceFile.IndexOf('\clips.dat')) { 
				$source_is_db_clips_dat_file = $true
			}

			# Delete Existing *.dat files ASYNCHRONOUSLY
			Get-Childitem -Path ($dest_path + "\*.dat") | Remove-Item -Force
			
			# Create a working copy of the Blue Iris database file
			$dat_name = Split-Path -leaf $SourceFile
			Copy-Item -LiteralPath $SourceFile -Destination $("$dest_path\$dat_name") -Force
			if (Test-Path -LiteralPath $SourceFile) {
				$source_path = $dest_path + "\" + $dat_name
				$sfx = "`nNOTE: to avoid conflicts, script created a working copy of -SourceFile"
			} else {
				$err_msg = "ERR/unable to create a working copy of '$SourceFile'"
				return $err_msg
			}
		} else {
			$source_path = $SourceFile
		}
	} else {
		$err_msg = "ERR/-SourceFile does not exist"
		return $err_msg
	}
} else {
	$err_msg = "ERR/missing required argument -SourceFile"
	return $err_msg
}

if (Test-Path -LiteralPath $source_path) {

	# INSPECT THE DAT FILE
	if ($FromConsole) {	
		Write-Host ("Source file: '$source_path'" + $sfx)
	}
	
	$binary_data = [System.IO.File]::ReadAllBytes($source_path)
	$binary_data_len = $binary_data.Length
	$progress_inc = if ($source_is_db_clips_dat_file) { 300 } else { 1 }
	$jpg_offsets = @()
	$jpgCount = 0
	
	if ($FromConsole) {	
		Write-Host ("Counting embedded JPG images (start byte FFD8, end byte FFD9) ") -NoNewLine
	}

	# Start performance timer
	$timer_inspect = [Diagnostics.Stopwatch]::StartNew()
	
	for ($i = 0; $i -lt $binary_data_len - 1; $i++) {
		if ($binary_data[$i] -eq 0xFF -and $binary_data[$i + 1] -eq 0xD8) {
			$jpg_offsets += $i
			if ($FromConsole) {	
				if ($i % $progress_inc -eq 0) {
					Write-Host (".") -NoNewLine
				}
			}
		}
	}

	# Stop overall performance timer
	$timer_inspect.stop()
	$total_ms = [Math]::Round($timer_inspect.elapsed.totalmilliseconds,1)
	$total_sec = $total_ms / 1000

	$jpg_offsets_cnt  = $jpg_offsets.Count 

	if ($FromConsole) {	
		Write-Host ""
		Write-Host("Found " + $jpg_offsets_cnt + " apparent images")
		if ($DevMode) {
			if ($total_ms -lt 9999) {
				$str = "Source file inspection time = {0:N2} msec" -f $total_ms
			} else {
				$str = "Source file inspection time = {0:N3} sec" -f $total_sec
			}
			Write-Host ("(dev mode) " + $str) -ForegroundColor Cyan
		}
	}

	# For debugging
	#Write-Output $jpg_offsets

	# EXTRACT EMBEDDED IMAGES IN THE DAT FILE
	if ($Extract) {
		
		if ($jpg_offsets_cnt -gt 0) {

			# Make -Range endpoints
			if ($Range.Count -eq 2) {
				$img_rng_ary = $Range  # user's requested limits
			} elseif ($Range.Count -eq 1) {  # Like -Range 5, -Range 0, or -Range -5
				
				# For debugging
				# Write-Host ("Range[0] = " + $Range[0] + ", type = " + $Range[0].GetType().Name)

				if ($Range[0] -gt 0) {
					$img_rng_ary = @(1, $Range[0])
				} elseif ($Range[0] -eq 0) {
					$img_rng_ary = @(1, $jpg_offsets_cnt)
				} else {
					$x = [Math]::Max(1, $jpg_offsets_cnt + $Range[0] + 1)
					$img_rng_ary = @($x, $jpg_offsets_cnt)
				}
			} else {
				$img_rng_ary = @(1, 1)  # Default
			}

			# For debugging
			# Write-Host ("img_rng_ary = " + ($img_rng_ary | ConvertTo-Json))
			
			# Validate -Range request
			if ($img_rng_ary[0] -eq 0 -or $img_rng_ary[1] -eq 0) {
				$err_msg = "ERR/malformed argument -Range x,y (x & y must be > 0)"
				return $err_msg
			}
			if ($img_rng_ary[1] -lt $img_rng_ary[0]) {
				$err_msg = "ERR/malformed argument -Range x,y (y must be >= x)"
				return $err_msg
			}
			if ($img_rng_ary[1] -gt $jpg_offsets_cnt) {
				$img_rng_ary[1] -gt $jpg_offsets_cnt
				if ($FromConsole) {	
					Write-Host "HEADS-UP - argument -Range exceeds the count of images in the DAT file"
				}
			}
			
			if ($FromConsole) {	
				Write-Host ("Extracting images " + $img_rng_ary[0] + " to " + $img_rng_ary[1] + " ") -NoNewLine
			}

			# Delete Existing *.jpg Files ASYNCHRONOUSLY
			Get-Childitem -Path ($extract_path + "\*.jpg") | Remove-Item -Force
		
			# Determine number of padded zeros for filename image number
			$pad_len = [Math]::Ceiling([Math]::Log10($jpg_offsets.Count))

			# Start performance timer
			$timer_extract = [Diagnostics.Stopwatch]::StartNew()
			
			$progress_inc = if ($source_is_db_clips_dat_file) { 100 } else { 1 }
			$jpg_cntr = 0
			$i = 0  # loop index
			foreach ($offset in $jpg_offsets) {

				# For debugging
				# Write-Host ""
				# Write-Host ("`noffset = " + "{0:X8}" -f $offset)
				
				$i++
				if ($i -lt $img_rng_ary[0]) {
					continue
				}
				if ($i -gt $img_rng_ary[1]) {
					break
				}
				
				# Show progress indicator
				if ($FromConsole) {	
					if ($i % $progress_inc -eq 0) {
						Write-Host (".") -NoNewLine
					}
				}
				
				$endIndex = $binary_data_len
				for ($j = $offset; $j -lt $binary_data_len - 1; $j++) {
					if ($binary_data[$j] -eq 0xFF -and $binary_data[$j + 1] -eq 0xD9) {  # JPG End Marker
						$endIndex = $j + 2
						break
					}
				}
				$jpg_cntr++
				
				# Save the image
				$jpg_filespec = $extract_path + "\" + "{0:d$pad_len}" -f $i + "." + "{0:X8}.jpg" -f $offset
				[System.IO.File]::WriteAllBytes($jpg_filespec, $binary_data[$offset..($endIndex-1)])

				# For debugging
				# Write-Host ""
				# Write-Host ("offset = " + "{0:X8}" -f $offset)
				# Write-Host "jpg_cntr = $jpg_cntr"
				
			}
			# Stop performance timer
			$timer_extract.stop()
			$total_ms = [Math]::Round($timer_extract.elapsed.totalmilliseconds,1)
			$total_sec = $total_ms / 1000

			if ($FromConsole) {	
				Write-Host ""
				Write-Host("Saved " + $jpg_cntr + " images to '" + $extract_path + "'")
				if ($DevMode) {
					if ($total_ms -lt 999) {
						$str = "JPG extraction time = {0:N2} msec" -f $total_ms
					} else {
						$str = "JPG extraction time = {0:N0} sec" -f $total_sec
					}
					#Write-Host ("(dev mode) " + $str) -ForegroundColor Cyan
					$avg = $total_ms/$jpg_cntr
					$avg_str = " ({0:N2} msec/image)" -f $avg
					Write-Host ("(dev mode) " + $str + $avg_str) -ForegroundColor Cyan
				}
			}
		} else {
			if ($FromConsole) {	
				Write-Host "No images found to extract"
			}
		}
	}

} else {
	$err_msg = "ERR/source file does not exist '" + $source_path + "'"
	return $err_msg
}

# HANDLE CLI MODE
if (!$FromConsole) {
	return $jpg_offsets_cnt
}


